#ifndef TOSTRINGS_H
#define TOSTRINGS_H

//#include "../lib/catch.hpp"

//#include <string>
//#include <QString>

//namespace Catch {
//template<> std::string toString<QString>(QString const& s);
//}

#endif // TOSTRINGS_H

